const standardOptions = {
    method: "POST",
    headers: {
        "Content-type": "application/json",
        Accept: "application/json"
    }
};
const randChannelName = () => Math.ceil(Math.random() * 1000000).toString();
export const noOpTag = (strings, ...exps) => {
    return strings.map((str, i) => str + (i in exps ? exps[i] : "")).join("");
};
export const gql = noOpTag;
export var GraphQLResponseCondition;
(function (GraphQLResponseCondition) {
    GraphQLResponseCondition["OK"] = "OK";
    GraphQLResponseCondition["InvalidRequest"] = "Invalid request";
    GraphQLResponseCondition["NetworkCallFailed"] = "Network call failed";
    GraphQLResponseCondition["InvalidResponse"] = "Invalid response";
    GraphQLResponseCondition["RequestError"] = "Request error";
    GraphQLResponseCondition["FieldError"] = "Field error";
})(GraphQLResponseCondition || (GraphQLResponseCondition = {}));
export class Api {
    constructor(url, apiOptions = {}) {
        const options = typeof apiOptions === "string"
            ? { wsUrl: apiOptions }
            : apiOptions;
        if (!options.fetch && typeof window === "undefined") {
            throw new Error("No fetch implementation provided and not operating in browser context");
        }
        options.fetch = options.fetch || window.fetch.bind(window);
        const protocol = url.match(/^(https?):\/\//)[1];
        if (!protocol)
            throw new Error(`Unexpected API URL [${url}]`);
        const isSecure = protocol.match(/s$/);
        this.url = url;
        this.wsUrl = options.wsUrl || [`ws${isSecure ? "s" : ""}:`, url.split("//").slice(1)].join("//");
        this.fetch = options.fetch;
        this.headers = options.headers;
        this.onReponse = options.onResponse;
    }
    get log() {
        return this.socket.log;
    }
    set log(fn) {
        if (typeof fn === "function") {
            if (this.socket)
                this.socket.log = fn;
        }
    }
    async run(query, variables = {}) {
        const graphQLResponse = (condition, { data, error, graphQLErrors }) => {
            const result = {
                ...response, data, condition, error, graphQLErrors, query, variables
            };
            if (typeof this.onReponse === "function") {
                this.onReponse(result);
            }
            return result;
        };
        let body;
        let response;
        try {
            body = JSON.stringify({ query, variables });
        }
        catch (error) {
            return graphQLResponse(GraphQLResponseCondition.InvalidRequest, { error });
        }
        try {
            response = await this.fetch(this.url, {
                ...standardOptions,
                headers: { ...standardOptions.headers, ...this.headers },
                body
            });
        }
        catch (error) {
            return graphQLResponse(GraphQLResponseCondition.NetworkCallFailed, { error });
        }
        let errors;
        let data;
        try {
            const responseJson = await response.clone().json();
            errors = responseJson.errors;
            data = responseJson.data;
        }
        catch (error) {
            return graphQLResponse(GraphQLResponseCondition.InvalidResponse, { error });
        }
        const condition = Array.isArray(errors) && errors.length
            ? [undefined, null].includes(data) ||
                (typeof data === "object" && Object.keys(data).every(key => [undefined, null].includes(data[key])))
                ? GraphQLResponseCondition.RequestError
                : GraphQLResponseCondition.FieldError
            : GraphQLResponseCondition.OK;
        return graphQLResponse(condition, { data, graphQLErrors: errors });
    }
    subscribe(query, variables = {}, channelName = randChannelName()) {
        const message = {
            id: channelName,
            type: "start",
            payload: {
                query,
                variables
            }
        };
        const startSub = () => this.socket.webSocket.send(JSON.stringify(message));
        if (this.socket) {
            if (this.socket.subscriptions[channelName]) {
                throw new Error(`Subscription already exists for channel [${channelName}]`);
            }
            if (this.socket.connected) {
                startSub();
            }
            else {
                this.socket.connectedHandlers.push(startSub);
            }
        }
        else {
            this.socket = new Socket(this.wsUrl, this.headers, startSub);
            if (this.log)
                this.socket.log = this.log;
        }
        const socketSubscription = this.socket.subscriptions[channelName] = {};
        const result = {
            onData(handler) {
                socketSubscription.dataHandler = handler;
                return result;
            },
            close() {
                delete this.socket.subscriptions[channelName];
            }
        };
        return result;
    }
}
class Socket {
    constructor(wsUrl, headers = {}, onConnected = () => { }) {
        this.connected = false;
        this.connectedHandlers = onConnected ? [onConnected] : [];
        this.subscriptions = {};
        this.webSocket = new WebSocket(wsUrl, "graphql-subscriptions");
        this.webSocket.onopen = () => {
            const message = {
                type: "connection_init",
                payload: headers
            };
            this.webSocket.send(JSON.stringify((message)));
        };
        this.webSocket.onmessage = event => {
            const data = JSON.parse(event.data);
            let msg;
            if (["subscription_fail", "error"].includes(data.type)) {
                const msg = `GraphQL ${data.type} for channel ${data.id} ${(data?.payload.message) || ""}`;
                throw new Error(msg);
            }
            else if (data.type === "data") {
                if (this.subscriptions[data.id] && typeof this.subscriptions[data.id].dataHandler === "function") {
                    return this.subscriptions[data.id].dataHandler(data.payload.data);
                }
                else
                    msg = `data received for channel [${data.id}] with no subscription handler`;
            }
            else if (data.type !== "ka") {
                const messages = {
                    connection_ack: `[${data.id}] connection_ack, the handshake is complete`,
                    init_fail: `[${data.id}] init_fail returned from the WebSocket server`,
                    subscription_success: `[${data.id}] subscription_success`
                };
                msg = messages[data.type];
                msg = msg || `unexpected message type [${data.type}] received from WebSocket server`;
                if (data.type === "connection_ack" && typeof onConnected === "function") {
                    this.connected = true;
                    this.connectedHandlers.forEach(connHandler => connHandler());
                }
            }
            if (msg && typeof this.log === "function")
                this.log(`graphqlx: ${msg}`);
        };
    }
}
export const trimInput = (proposedInput, typeName, inputTypeInfoMap) => {
    const inputType = inputTypeInfoMap[typeName];
    if (!inputType)
        return proposedInput;
    const result = {};
    for (const { name, typeName, isList } of inputType.fields) {
        const fieldName = name;
        const field = proposedInput[fieldName];
        if (typeof field !== "undefined") {
            if (field !== null && typeName in inputTypeInfoMap) {
                if (isList && Array.isArray(field)) {
                    const list = [];
                    for (const item of field) {
                        list.push(trimInput(item, typeName, inputTypeInfoMap));
                    }
                    result[fieldName] = list;
                }
                else {
                    result[fieldName] = trimInput(field, typeName, inputTypeInfoMap);
                }
            }
            else {
                result[fieldName] = field;
            }
        }
    }
    return result;
};
//# sourceMappingURL=graphqlex.js.map